@@ -1,26 +1,34 @@
-from cmath import isfinite
-from dataclasses import dataclass
-
-from typing import List, Optional, Tuple, Type, Union, Any, Dict, Callable
-import numpy as np
-import tensorflow as tf
 import gc
-from yaket.schema.schema import TrainingModel, yaml_to_pydantic, Accelerator
-from yaket.converter.converter import Converter
 import importlib
-import mlflow
 import os
-import time
 import subprocess as sp
 import sys
+import time
+from dataclasses import dataclass
+from typing import Any, Callable, Dict, List, Optional, Tuple, Union
 
+import mlflow
+import numpy as np
+import tensorflow as tf
+
+from yaket.converter.converter import Converter
+from yaket.schema.schema import Accelerator, TrainingModel, yaml_to_pydantic
+import platform
 
 @dataclass
 class Trainer:
     config_params: Union[Dict, str]
     model: tf.keras.Model
-    train_dataset: Union[Tuple[np.ndarray, np.ndarray],Tuple[np.ndarray, np.ndarray, np.ndarray], tf.data.Dataset]
-    val_dataset: Union[Tuple[np.ndarray, np.ndarray],Tuple[np.ndarray, np.ndarray, np.ndarray], tf.data.Dataset] = None
+    train_dataset: Union[
+        Tuple[np.ndarray, np.ndarray],
+        Tuple[np.ndarray, np.ndarray, np.ndarray],
+        tf.data.Dataset,
+    ]
+    val_dataset: Union[
+        Tuple[np.ndarray, np.ndarray],
+        Tuple[np.ndarray, np.ndarray, np.ndarray],
+        tf.data.Dataset,
+    ] = None
     strategy: Optional[tf.distribute.Strategy] = None
     random_seed: int = 1234
     validate_yaml: bool = True
@@ -42,51 +50,36 @@ class Trainer:
     _user_strategy: bool = False
     _trainer_initialized: bool = False
     _sample_weight_mode: Optional[str] = None
-    
 
-    def __post_init__(self):
-        """Initialize the trainer: check inputs, load custom modules, parse configuration file."""
+    # __post_init__ is called after the __init__ function automatically
 
-        if not isinstance(self.model, tf.keras.models.Model):
-            raise Exception("model must be keras model")
-        if isinstance(self.train_dataset, tuple):
-            if len(self.train_dataset) <2 or len(self.train_dataset) > 3:
-                raise Exception("train_dataset must be a tuple of (x, y) or (x, y, sample_weight)")
-        if isinstance(self.val_dataset, tuple):
-            if len(self.val_dataset) < 2 or len(self.val_dataset) > 3:
-                raise Exception("val_dataset must be a tuple of (x, y) or (x, y, sample_weight)")
-        if not isinstance(self.config_params, str) and not isinstance(self.config_params, dict):
-            raise Exception("Config_params must be a dictionary or a file path")
-        if isinstance(self.config_params, str) and not os.path.isfile(
-            self.config_params
-        ):
-            raise Exception("Config_params must be a valid file path")
+    def train(self, epochs: int = None):
+        """Train the model. Main function to call.
 
-        if self.strategy is not None and not isinstance(
-            self.strategy, tf.distribute.Strategy
-        ):
-            raise Exception("Strategy must be keras strategy object")
-        if not isinstance(self.random_seed, int):
-            raise Exception("Random seed must be an integer")
-        if not isinstance(self.validate_yaml, bool):
-            raise Exception("Validate yaml must be a boolean value")
-        if self.custom_modules_path is not None:
-            if not isinstance(self.custom_modules_path, str) or not os.path.isfile(
-                self.custom_modules_path
-            ):
-                raise Exception("Costum modules path must be a valid path string")
+        Args
+        ----
+        epochs: int, optional
+            Number of epochs to train. If None, will train with the number of epochs specified in the config file.
 
-        if self.custom_modules_path:
-            self._import_custom_model(self.custom_modules_path)
-        self._config = self._parse_config()
-        self._accelerator = Accelerator[self.config.accelerator]
-        self._validate_config_file()
+        """
 
-        self._callbacks = self._get_callbacks()
-        self._input_shape = self._get_input_shape()
-        self._sample_weight_mode = self._get_sample_weight_mode()
+        if self.strategy is None and self._accelerator is Accelerator.cpu:
+            train_dataset = self._get_x_y_train(self.config.batch_size)
+            val_dataset = self._get_x_y_val(self.config.batch_size)
+            history = self._train(train_dataset, val_dataset, epochs)
+        else:
+            strategy = self._get_strategy()
+            batch_size = strategy.num_replicas_in_sync * self.config.batch_size
+            train_dataset = self._get_x_y_train(batch_size)
+            val_dataset = self._get_x_y_val(batch_size)
+            with strategy.scope():
+                self._clone_model() if not self._user_strategy else None  # Clone model only if not using user strategy (e.g. CustoModule)
+                history = self._train(train_dataset, val_dataset, epochs)
 
-        self._trainer_initialized = True
+        self._save_model()
+        self._history = history.history
+        self._clean_workspace()
+        return history
 
     def _train(self, train_dataset, val_dataset, epochs: int = None):
         """Train the model"""
@@ -99,40 +92,16 @@ class Trainer:
             validation_data=val_dataset,
             batch_size=None,
             callbacks=self._callbacks,
-            steps_per_epoch=int(self.config.steps_per_epoch) if self.config.steps_per_epoch is not None else None,
+            steps_per_epoch=int(self.config.steps_per_epoch)
+            if self.config.steps_per_epoch is not None
+            else None,
             class_weight=None,  # TODO: add class_weight,
             verbose=int(self.config.verbose),
         )
         return history
 
-    def train(self, epochs: int = None):
-        """Train the model. Main function to call.
+    
 
-        Args
-        ----
-        epochs: int, optional
-            Number of epochs to train. If None, will train with the number of epochs specified in the config file.
-
-        """
-
-
-        if self.strategy is None and self._accelerator is Accelerator.cpu:
-            train_dataset = self._get_x_y_train(self.config.batch_size)
-            val_dataset = self._get_x_y_val(self.config.batch_size)
-            history = self._train(train_dataset, val_dataset, epochs)
-        else:
-            strategy = self._get_strategy()
-            batch_size = strategy.num_replicas_in_sync * self.config.batch_size
-            train_dataset = self._get_x_y_train(batch_size)
-            val_dataset = self._get_x_y_val(batch_size)
-            with strategy.scope():
-                self._clone_model() if not self._user_strategy else None # Clone model only if not using user strategy (e.g. CustoModule)
-                history = self._train(train_dataset, val_dataset, epochs)
-
-        self._save_model()
-        self._history = history.history
-        self._clean_workspace()
-        return history
 
     def convert_model(
         self,
@@ -148,7 +117,7 @@ class Trainer:
 
         Args
         ----
-        format: str
+        format_model: str
             Format to convert the model to. Available formats: onnx, tflite
         model_path: str
             Path to the model to convert.
@@ -159,26 +128,35 @@ class Trainer:
             It might be not available with some platforms.
 
         """
+        if not isinstance(format_model, str):
+            raise TypeError("format_model must be a string")
+        if not isinstance(output_path, str):
+            raise TypeError("output_path must be a string")
+        if not isinstance(from_command_line, bool):
+            raise TypeError("from_command_line must be a boolean")
+        if not isinstance(opset_onnx, int) or (isinstance(opset_onnx, int) and opset_onnx < 1):
+            raise TypeError("opset_onnx must be a positive integer")
+
         if from_command_line and self._out_path is None:
-            raise Exception("You need to have a saved model before converting using command line")
+            raise Exception(
+                "You need to have a saved model before converting using command line"
+            )
         model_to_convert = self.model if not from_command_line else None
         converter = Converter(
             out_format=format_model.lower(),
             opset_onnx=opset_onnx,
             model_path=self._out_path,
             model=model_to_convert,
-            out_path = output_path,
+            out_path=output_path,
         )
         if converter.convert():
             print(f"Successfully converted to format {format} ")
 
-    
     def _get_sample_weight_mode(self):
         """Get the sample weight mode from the config file"""
         if self.config.sample_weight_mode is not None:
             return self.config.sample_weight_mode
 
-
     def _get_input_shape(self):
         """Get the input shape of input dataset"""
         if isinstance(self.train_dataset, tf.data.Dataset):
@@ -190,8 +168,6 @@ class Trainer:
 
     def _save_model(self):
         """Save the model by loading best checkpoint if available and saving it to mlflow or local path"""
-        #TODO: add custom model saving with weights
-
 
         if self._model_checkpoint is not None:
             self.model.load_weights(self._model_checkpoint)
@@ -208,14 +184,11 @@ class Trainer:
             os.makedirs(os.getcwd() + "/models", exist_ok=True)
             t = int(time.time())
             self._out_path = os.getcwd() + f"/models/{t}_best_model"
-        
+
         try:
-            self.model.save(self._out_path) # if not custom model/layers
+            self.model.save(self._out_path)  # if not custom model/layers
         except:
-            self.model.save_weights(self._out_path, save_format='tf')
-
-            
-
+            self.model.save_weights(self._out_path, save_format="tf")
 
     def _clone_model(self):
         """Clone the model so that it works within tf.distribute.Strategy
@@ -223,13 +196,13 @@ class Trainer:
         """
 
         if not self.model._is_graph_network:
-            raise Exception("Model must be be a Sequential or Functional API model without custom objects")
+            raise Exception(
+                "Model must be be a Sequential or Functional API model without custom objects"
+            )
         try:
             self.model = tf.keras.models.clone_model(self.model)
-        except NotImplementedError: # there's custom layer
-            raise 
-    
-
+        except NotImplementedError:  # there's custom layer
+            raise
 
     def _get_x_y_val(self, batch_size):
         """Get the x and y for training based on the format of the dataset"""
@@ -269,17 +242,15 @@ class Trainer:
     def validate_config(self):
         """Validate again the configuration file. Used after chaning the config parameters"""
 
-        temp_dir = {k:v for k,v in self.config.__dict__.items() if v is not None}
+        temp_dir = {k: v for k, v in self.config.__dict__.items() if v is not None}
 
         TrainingModel(**temp_dir)
         self._validate_config_file()
 
-
     @property
     def config(self):
         return self._config
 
-
     def _compile_model(self) -> None:
         """Compile the model"""
 
@@ -288,7 +259,10 @@ class Trainer:
         self._metrics = self._get_metrics()
 
         self.model.compile(
-            optimizer=self._optimizer, loss=self._loss, metrics=self._get_metrics(), sample_weight_mode = self._sample_weight_mode
+            optimizer=self._optimizer,
+            loss=self._loss,
+            metrics=self._get_metrics(),
+            sample_weight_mode=self._sample_weight_mode,
         )
 
     def _get_strategy(self):
@@ -314,6 +288,12 @@ class Trainer:
     @staticmethod
     def get_free_gpu_idx():
         """Get the index of the freer GPU"""
+        if platform.system == "Windows":
+            raise Exception("GPU not supported on Windows")
+        gpus = tf.config.experimental.list_physical_devices("GPU")
+        if len(gpus) == 0:
+            raise Exception("No GPU available")
+
         command = "nvidia-smi --query-gpu=memory.free --format=csv"
         memory_free_info = (
             sp.check_output(command.split()).decode("ascii").split("\n")[:-1][1:]
@@ -327,7 +307,6 @@ class Trainer:
         if isinstance(self.config_params, dict):
             return TrainingModel(**self.config_params)
 
-
     def _validate_config_file(self):
         "Validate existence of the loss, optimizer and callbacks defined in the config file"
         try:
@@ -335,7 +314,7 @@ class Trainer:
             self._get_metrics()
             self._get_loss()
         except Exception as e:
-            raise TypeError(
+            raise ValueError(
                 f"You are using a module not defined in either keras or in the custom script\n Details: {e}"
             )
 
@@ -366,18 +345,16 @@ class Trainer:
         opt = self.config.optimizer
         if isinstance(opt, List):
             if len(opt) == 1 and isinstance(opt[0], Dict):
-                k = list(opt[1].keys())[0]
-                v = list(opt[1].values())[0]
-                scheduler = getattr(tf.keras.optimizers, k, default_value)
+                k = list(opt[0].keys())[0]
+                v = list(opt[0].values())[0]
+                optimizer = getattr(tf.keras.optimizers, k, default_value)
                 if optimizer != default_value:
                     return optimizer(**opt_pars)
                 else:
                     return self._load_custom_module(optimizer, opt_pars)
             else:
                 opt_name = opt[0]
-                optimizer = getattr(
-                    tf.keras.optimizers, f"{opt_name}", default_value
-                )
+                optimizer = getattr(tf.keras.optimizers, f"{opt_name}", default_value)
                 if optimizer == default_value:
                     optimizer = self._load_custom_module(optimizer)
                 if isinstance(opt[1], dict):
@@ -385,18 +362,20 @@ class Trainer:
                     v = list(opt[1].values())[0]
                     scheduler = getattr(tf.keras.optimizers.schedules, k, default_value)
                     if scheduler == default_value:
-                        raise ValueError(f"{v} is not a valid scheduler. Only available from keras")
+                        raise ValueError(
+                            f"{v} is not a valid scheduler. Only available from keras"
+                        )
                     else:
-                        opt_pars['learning_rate'] = scheduler(**v)
+                        opt_pars["learning_rate"] = scheduler(**v)
                         return optimizer(**opt_pars)
         elif isinstance(opt, str):
             optimizer = getattr(
-            tf.keras.optimizers, f"{self.config.optimizer}", default_value
+                tf.keras.optimizers, f"{self.config.optimizer}", default_value
             )
             if optimizer != default_value:
                 return optimizer(**opt_pars)
             else:
-                    return self._load_custom_module(optimizer, opt_pars)
+                return self._load_custom_module(optimizer, opt_pars)
 
     def _get_loss(self) -> Union[tf.keras.losses.Loss, Callable]:
         """Get the loss from the config file"""
@@ -419,7 +398,6 @@ class Trainer:
             else:
                 return self._load_custom_module(loss_config, loss_params)
 
-
     def _get_callbacks(self) -> List[tf.keras.callbacks.Callback]:
         """Get the callbacks from the config file"""
         if self.config.callbacks is None:
@@ -510,9 +488,11 @@ class Trainer:
 
     def _set_randomness(self, random_seed: Optional[int] = None) -> None:
         """Set the randomness"""
+        if (not isinstance(random_seed, int)) or (isinstance(random_seed, int) and random_seed < 0):
+            raise ValueError("random_seed must be a positive integer")
         if random_seed is not None:
             if tf.__version__ >= "2.9.0":
-                tf.keras.set_random_seed(random_seed)
+                tf.random.set_seed(random_seed)
                 tf.config.experimental.enable_op_determinism()
             else:
                 tf.random.set_seed(random_seed)
@@ -530,14 +510,81 @@ class Trainer:
 
     def summary_model(self):
         """Summary of the model"""
+        if self.model is None:
+            raise ValueError("Model not found. Please initialize the Trainer first")
         self.model.summary()
 
+    def __post_init__(self):
+        """Initialize the trainer: check inputs, load custom modules, parse configuration file."""
+
+        if not isinstance(self.model, tf.keras.models.Model):
+            raise TypeError("model must be keras model")
+        if isinstance(self.train_dataset, tuple):
+            if len(self.train_dataset) < 2 or len(self.train_dataset) > 3:
+                raise ValueError(
+                    "train_dataset must be a tuple of (x, y) or (x, y, sample_weight)"
+                )
+            for val in self.train_dataset:
+                if not isinstance(val, np.ndarray):
+                    raise TypeError("train_dataset must be a tuple of numpy arrays")
+        if isinstance(self.val_dataset, tuple):
+            if len(self.val_dataset) < 2 or len(self.val_dataset) > 3:
+                raise ValueError(
+                    "val_dataset must be a tuple of (x, y) or (x, y, sample_weight)"
+                )
+            for val in self.val_dataset:
+                if not isinstance(val, np.ndarray):
+                    raise TypeError("val_dataset must be a tuple of numpy arrays")
+        if not (
+            isinstance(self.config_params, dict) or isinstance(self.config_params, str)
+        ):
+            raise TypeError("Config_params must be a dictionary or a file path")
+        if isinstance(self.config_params, str) and not os.path.isfile(
+            self.config_params
+        ):
+            raise FileNotFoundError("Config_params must be a valid file path")
+
+        if self.strategy is not None and not isinstance(
+            self.strategy, tf.distribute.Strategy
+        ):
+            raise TypeError("Strategy must be keras strategy object")
+        if not isinstance(self.random_seed, int):
+            raise TypeError("Random seed must be an integer")
+        if not isinstance(self.validate_yaml, bool):
+            raise TypeError("Validate yaml must be a boolean value")
+        if self.custom_modules_path is not None:
+            if not isinstance(self.custom_modules_path, str) or not os.path.isfile(
+                self.custom_modules_path
+            ):
+                raise FileNotFoundError(
+                    "Costum modules path must be a valid path string"
+                )
+
+        if self.custom_modules_path:
+            self._import_custom_model(self.custom_modules_path)
+        self._config = self._parse_config()
+        self._accelerator = Accelerator[self.config.accelerator]
+        self._validate_config_file()
+
+        self._callbacks = self._get_callbacks()
+        self._input_shape = self._get_input_shape()
+        self._sample_weight_mode = self._get_sample_weight_mode()
+
+        self._trainer_initialized = True
+
 
 if __name__ == "__main__":
     import numpy as np
     from tensorflow import keras
     from tensorflow.keras import layers
 
+    model = keras.models.Sequential()
+    model.add(layers.Dense(64, activation="relu"))
+
+    trainer_path = "/root/project/yaket/examples/files/03_trainer.yaml"
+    trainer = Trainer(
+        trainer_path, model, train_dataset=(None, None), val_dataset=(None, None)
+    )
 
     class MyDenseLayer(tf.keras.layers.Layer):
         def __init__(self, num_outputs):
@@ -545,9 +592,9 @@ if __name__ == "__main__":
             self.num_outputs = num_outputs
 
         def build(self, input_shape):
-            self.kernel = self.add_weight("kernel",
-                                        shape=[int(input_shape[-1]),
-                                                self.num_outputs])
+            self.kernel = self.add_weight(
+                "kernel", shape=[int(input_shape[-1]), self.num_outputs]
+            )
 
         def call(self, inputs):
             return tf.matmul(inputs, self.kernel)
@@ -590,20 +637,21 @@ if __name__ == "__main__":
 
     path = "/root/project/yaket/examples/files/trainer.yaml"
 
-    from sklearn.utils.class_weight import compute_sample_weight
     from sklearn.utils import class_weight
+    from sklearn.utils.class_weight import compute_sample_weight
 
     # define your class weights or automatically compute them
 
     yargmax = np.argmax(y_train, axis=1)
-    class_weights = class_weight.compute_class_weight('balanced',
-                                                    classes = np.unique(yargmax.flatten()),
-                                                    y = yargmax.flatten())
+    class_weights = class_weight.compute_class_weight(
+        "balanced", classes=np.unique(yargmax.flatten()), y=yargmax.flatten()
+    )
     class_weight_dict = dict(enumerate(class_weights))
 
-
     # compute the sample weights
-    sample_weigth = compute_sample_weight(class_weight_dict, yargmax.flatten()).reshape(yargmax.shape)
+    sample_weigth = compute_sample_weight(class_weight_dict, yargmax.flatten()).reshape(
+        yargmax.shape
+    )
     print(y_train.shape, sample_weigth.shape, x_train.shape)
 
     trainer = Trainer(
